#include "stdafx.h"
#include "mdk.h"
#include "../dsplib/dsplib.h"

#define MDK_VERSION		1

class CInput
{
public:
	CInput(char const *n, bool st) : Name(n), Stereo(st) {}

public:
	std::string Name;
	bool Stereo;

};

typedef std::list<CInput> InputList;
typedef std::list<CInput>::iterator InputListIterator;


void CopyStereoToMono(float *pout, float *pin, int numsamples, float amp)
{
	do
	{
		*pout++ = (pin[0] + pin[1]) * amp;
		pin += 2;
	} while(--numsamples);
}

void AddStereoToMono(float *pout, float *pin, int numsamples, float amp)
{
	do
	{
		*pout++ += (pin[0] + pin[1]) * amp;
		pin += 2;
	} while(--numsamples);
}

void CopyM2S(float *pout, float *pin, int numsamples, float amp)
{
	do
	{
		double s = *pin++ * amp;
		pout[0] = (float)s;
		pout[1] = (float)s;
		pout += 2;
	} while(--numsamples);

}

void Add(float *pout, float *pin, int numsamples, float amp)
{
	do
	{
		*pout++ += *pin++ * amp;
	} while(--numsamples);
}

class CMDKImplementation
{
	friend class CMDKMachineInterface;
	friend class CMDKMachineInterfaceEx;

	void AddInput(char const *macname, bool stereo)
	{
		if (macname == NULL)
			return;
 
		Inputs.push_back(CInput(macname, stereo));


		SetMode();
	}

	void DeleteInput(char const *macname)
	{
		for (InputList::iterator i = Inputs.begin(); i != Inputs.end(); i++)
		{
			if ((*i).Name.compare(macname) == 0)
			{
				Inputs.erase(i);
				SetMode();
				return;
			}
		}
	}

	void RenameInput(char const *macoldname, char const *macnewname)
	{
		for (InputList::iterator i = Inputs.begin(); i != Inputs.end(); i++)
		{
			if ((*i).Name.compare(macoldname) == 0)
			{
				(*i).Name = macnewname;
				return;
			}
		}
	}

	void SetInputChannels(char const *macname, bool stereo)
	{
		for (InputList::iterator i = Inputs.begin(); i != Inputs.end(); i++)
		{
			if ((*i).Name.compare(macname) == 0)
			{
				(*i).Stereo = stereo;
				SetMode();
				return;
			}
		}
	}

	void Input(float *psamples, int numsamples, float amp)
	{
		assert(InputIterator != Inputs.end());

		if (psamples == NULL)
		{
			InputIterator++;
			return;
		}


		if (numChannels == 1)
		{
			if (HaveInput == 0)
			{
				if ((*InputIterator).Stereo)
					CopyStereoToMono(Buffer, psamples, numsamples, amp);
				else
					DSP_Copy(Buffer, psamples, numsamples, amp);
			}
			else
			{
				if ((*InputIterator).Stereo)
					AddStereoToMono(Buffer, psamples, numsamples, amp);
				else
				    DSP_Add(Buffer, psamples, numsamples, amp);
			}
		}
		else
		{
			if (HaveInput == 0)
			{
				if ((*InputIterator).Stereo)
					DSP_Copy(Buffer, psamples, numsamples*2, amp);
				else
					CopyM2S(Buffer, psamples, numsamples, amp);
			}
			else
			{
				if ((*InputIterator).Stereo) 
					DSP_Add(Buffer, psamples, numsamples*2, amp);
				else
					DSP_AddM2S(Buffer, psamples, numsamples, amp);
			}
		}

		HaveInput++;
		InputIterator++;
	
	}

	bool Work(float *psamples, int numsamples, int const mode)
	{
		if ((mode & WM_READ) && HaveInput)
			DSP_Copy(psamples, Buffer, numsamples);

		bool ret = pmi->MDKWork(psamples, numsamples, mode);

		InputIterator = Inputs.begin();
		HaveInput = 0;

		return ret;
	}

	bool WorkMonoToStereo(float *pin, float *pout, int numsamples, int const mode)
	{
		if ((mode & WM_READ) && HaveInput)
			DSP_Copy(pout, Buffer, 2*numsamples);

		bool ret = pmi->MDKWorkStereo(pout, numsamples, mode);

		InputIterator = Inputs.begin();
		HaveInput = 0;
		
		return ret;
	}

	
	void Init(CMachineDataInput * const pi)
	{
		ThisMachine = pmi->pCB->GetThisMachine();
		
		numChannels = 1;
	
		InputIterator = Inputs.begin();
		HaveInput = 0;
		MachineWantsChannels = 1;

		if (pi != NULL)
		{
			byte ver;
			pi->Read(ver);
		}
		

		pmi->MDKInit(pi);
	}

	void Save(CMachineDataOutput * const po)
	{
		po->Write((byte)MDK_VERSION);

		pmi->MDKSave(po);
	}

	void SetOutputMode(bool stereo)
	{
		numChannels = stereo ? 2 : 1;
		MachineWantsChannels = numChannels;
		
		pmi->OutputModeChanged(stereo);
	}

	void SetMode()
	{	
		InputIterator = Inputs.begin();
		HaveInput = 0;
		
		if (MachineWantsChannels > 1)
		{
			numChannels = MachineWantsChannels;
			pmi->pCB->SetnumOutputChannels(ThisMachine, numChannels);
			pmi->OutputModeChanged(numChannels > 1 ? true : false);
			return;
		}
	

		for (InputListIterator i = Inputs.begin(); i != Inputs.end(); i++)
		{
			if ((*i).Stereo)
			{
				numChannels = 2;
				pmi->pCB->SetnumOutputChannels(ThisMachine, numChannels);
				pmi->OutputModeChanged(numChannels > 1 ? true : false);
				return;
			}
		}

		numChannels = 1;
		pmi->pCB->SetnumOutputChannels(ThisMachine, numChannels);
		pmi->OutputModeChanged(numChannels > 1 ? true : false);
	
	}


public:
	CMDKMachineInterface *pmi;

	InputList Inputs;
	InputListIterator InputIterator;

	int HaveInput;
	int numChannels;
	int MachineWantsChannels;

	CMachine *ThisMachine;
	
	float Buffer[2*MAX_BUFFER_LENGTH];

};

void CMDKMachineInterfaceEx::AddInput(char const *macname, bool stereo) { pImp->AddInput(macname, stereo); }
void CMDKMachineInterfaceEx::DeleteInput(char const *macename) { pImp->DeleteInput(macename); }
void CMDKMachineInterfaceEx::RenameInput(char const *macoldname, char const *macnewname) { pImp->RenameInput(macoldname, macnewname); }
void CMDKMachineInterfaceEx::SetInputChannels(char const *macname, bool stereo) { pImp->SetInputChannels(macname, stereo); }
void CMDKMachineInterfaceEx::Input(float *psamples, int numsamples, float amp) { pImp->Input(psamples, numsamples, amp); }

bool CMDKMachineInterface::Work(float *psamples, int numsamples, int const mode) { return pImp->Work(psamples, numsamples, mode); }
bool CMDKMachineInterface::WorkMonoToStereo(float *pin, float *pout, int numsamples, int const mode) { return pImp->WorkMonoToStereo(pin, pout, numsamples, mode); }
void CMDKMachineInterface::Save(CMachineDataOutput * const po) { pImp->Save(po); }

void CMDKMachineInterface::SetOutputMode(bool stereo) { pImp->SetOutputMode(stereo); }


CMDKMachineInterface::~CMDKMachineInterface()
{
	delete pImp;
}

void CMDKMachineInterface::Init(CMachineDataInput * const pi)
{
	pImp = new CMDKImplementation;
	pImp->pmi = this;

	CMDKMachineInterfaceEx *pex = GetEx();
	pex->pImp = pImp;
	pCB->SetMachineInterfaceEx(pex);

	pImp->Init(pi);
}

